import re
import sets
import os.path

import utils


fof_re=re.compile('^\s*fof\s*\(\s*([^,\s]*)\s*,\s*([^,\s]*)\s*,(.*)\).')
fof_include=re.compile("^\s*include\s*\(\s*'([^']*)'\s*\).")
symbol_re=re.compile('(?<!\w)[_a-z][_a-zA-Z0-9]*')

def get_symbols(string):
	syms=symbol_re.findall(string)
	return sets.Set(syms)
	
class EarlyProblemAccessError(Exception):
	pass
class InvalidFormulaError(Exception):
	pass
class InvalidRefutationStringError(Exception):
	pass

class AxiomSet(object):
	file_prefix=None
	
	def __init__(self, fname):
		if self.__class__.file_prefix:
			fname=os.path.join(self.__class__.file_prefix, fname)
		self.fname=fname
		self._size=None
		
	def get_axioms(self):
		return load_formulas(self.fname)
		
	@property
	def size(self):
		if not self._size:
			self._size=os.path.getsize(self.fname)
		return self._size
	def __eq__(self, other):
		return hasattr(other, 'fname') and self.fname==other.fname
	def __hash__(self):
		return hash(self.fname)
	def __str__(self):
		return os.path.basename(self.fname)
	def __repr__(self):
		return os.path.basename(self.fname)

class AxiomSetSet(utils.HashableList):
	def __init__(self):
		self._size=None
		self.proof_axioms=set()
		self._axioms=None
		
	def get_axioms(self):
		if not self._axioms:
			res={}
			for axs in self:
				axd=axs.get_axioms()
				if not res:
					res=axd
				else:
					res.update(axd)
			self._axioms=res
		return self._axioms
	
	_all_objs={}
	@classmethod
	def normalize(cls, obj):
		'''Ensures that only one object exists in an equality class.'''
		if obj in cls._all_objs:
			return cls._all_objs[obj]
		else:
			cls._all_objs[obj]=obj
		return obj
	
	@property
	def size(self):
		if not self._size:
			self._size=0
			for axs in self:
				self._size+=axs.size
		return self._size
		
	def __eq__(self, other):
		if (not hasattr(other, 'proof_axioms')) or len(self)!=len(other):
			return False
		for el in self:
			if not (el in other):
				return False
		return True
		
class Problem(object):
	def __init__(self, prob_fname, out_fname=None):
		self.prob_fname=prob_fname
		self.out_fname=out_fname
		self._axiom_sets=None
		self._own_formulas=None
		self.current_phase=0
		self.prover_process=None
		self.active=False
	
	@property
	def own_formulas(self):
		if not self.active:
			raise EarlyProblemAccessError()
		if not self._own_formulas:
			self._own_formulas=load_formulas(self.prob_fname)
		return self._own_formulas
	
	@property
	def axiom_set_set(self):
		return self.axiom_sets

	@property
	def axiom_sets(self):
		if not self.active:
			raise EarlyProblemAccessError()
		if not self._axiom_sets:
			ax=AxiomSetSet()
			incl_re=re.compile("^include\('([^']+)'\)\.$")

			inp = open(self.prob_fname)
			try:
				for line in inp:
					match_obj=incl_re.match(line)
					if match_obj:
						ax.append(AxiomSet(match_obj.group(1)))
			finally:
				inp.close()
			ax.sort()
			
			self._axiom_sets=AxiomSetSet.normalize(ax)
		return self._axiom_sets
	
	def __eq__(self, other):
		return hasattr(other, 'out_fname') and self.out_fname==other.out_fname
	def __hash__(self):
		return hash(self.out_fname)
	def __str__(self):
		return os.path.basename(self.prob_fname)
	def __repr__(self):
		return os.path.basename(self.prob_fname)

class Formula(object):
	def __init__(self, line):
		self.full=line
		match_obj=fof_re.match(line)
		if not match_obj:
			raise InvalidFormulaError, line
		self.name=match_obj.group(1)
		self.type=match_obj.group(2)
		body=match_obj.group(3)
		self.symbols=get_symbols(body)
	
	def __eq__(self, other):
		return hasattr(other, 'name') and self.name==other.name
	def __hash__(self):
		return hash(self.name)
	def __str__(self):
		return self.full


#needs to be set to some temporary directory
fname_dir=None

def static_num2():
	k = 0
	while True:
		k += 1
		yield k
fname_index = static_num2().next

def get_fname(problem):
	pr_bname=os.path.basename(problem.prob_fname)
	return '%s/sel_%s_%d' % (fname_dir, pr_bname, fname_index())

def write_forms_to_file(problem, forms):
	fname=get_fname(problem)
	out = open(fname, 'w')
	try:
		lines=[f.full+'\n' for f in forms]
		out.writelines(lines)
	finally:
		out.close()
	return fname
	
		
def parenthesis_balance(str):
	res=0
	for c in str:
		if c=='(':
			res+=1
		elif c==')':
			res-=1
	return res

def load_formulas(fname):
	'''Loads formulas from specified file and returns them in associative array 
	<formula name>-><Formula object>.'''
	formulas={}
	depth=0
	form_lines=[]
	
	inp = open(fname)
	try:
		for line in inp:
			line=line.strip()
			line=line.split('%')[0]
			if not line:
				continue
			
			form_lines.append(line)
			depth+=parenthesis_balance(line)
			if depth==0:
				form_joined=reduce(lambda x,y: '%s %s'%(x,y), form_lines)
				
				try:
					form=Formula(form_joined)
					formulas[form.name]=form
				except InvalidFormulaError:
					match_obj=fof_include.match(form_joined)
					if match_obj:
						#We ignore includes when looking for formulas. 
						pass
					else:
						raise
				form_lines=[]
			elif depth<0:
				raise Exception
	finally:
		inp.close()
	return formulas
			
def get_proof_symbols(fname):
	refutation_re=re.compile('^(cnf|fof)\(\d+\s*,\s*(\w+)\s*,\s*(.*)\)\.$');
	symbols=sets.Set()
	
	inp = open(fname)
	try:
		for line in inp:
			line=line.strip()
			mo=refutation_re.match(line)
			if not mo:
				continue
			
			kind=mo.group(2)
			
			if kind!='axiom':
				continue
			
			inner_str=mo.group(3) #this contains the formula and how it was inferred
			
			depth=0
			index=0
			for c in inner_str:
				if c=='(':
					depth+=1
				elif c==')':
					depth-=1
				if c==',' and depth==0:
					break
				index+=1
			else:
				raise InvalidRefutationStringError(inner_str)
			
			formula=inner_str[0:index]
			inf_spec=inner_str[index+1:]
			
			if inf_spec.find('kb')==-1:
				continue
			
			form_syms=get_symbols(formula)
			symbols.update(form_syms)
	finally:
		inp.close()
	return symbols
	
